# get_rand.R, R.Nachtrieb, 2021-03-16
# accepts MISO data as argument
# calculates grand average of Demand, Wind, Solar and makes normalized variables
# construct normalized variables: subtract mean, divide by stdev
# calculates the factor means of the normalized variables

##
# we'll do this all with dplyr 
# (1) like to retrieve hourly signal (signal is one of demand, wind, solar)
# (2) rename hour of day (hod) and month-of-year (moy) columns
# (3) add hour-of-week (how) column
# (4) create normalized signal zz by subtracting overall mean and dividing by overall stdev
# (5) find the periodic signal pp by mean(zz) grouping by how, moy
# (6.) find the random hourly signal by subtracting pp from zz

library(psd)
library(tidyverse)
library(spectral)


# for k*x < 1, curve is approximately constant at y=a
# for k*x > 1, the curve decreases like 1/x^2
lorentz <- function(a=1,k=.001,x,s=2){
  y <- a/(1+(x/k)^s)
  return(y)
}

# returns normalized power law function
# scales ff to run 1..N/2
powerfun <- function(alpha=1,ff){
  nmax <- max(ff)/min(ff)/2
  xx <- ff/min(ff)
  if (alpha == 1) {
    p0inv <- log(nmax)
  } else {
    p0inv <- 1/(1-alpha)*(nmax^(1-alpha)-1)
  }
  pp <- xx^(-alpha)/p0inv
  return(pp)
}

##################################
# plots the power spectral distribution of a time series using the fourier transform method
#
# improvement: replace NA data with average of either side; should introduce no bias, 
# but preserve inter-temporal relationships

plot_fft <- function(
  xx=xx,tt=tt, 
  mainstr="",
  xcal=TRUE,
  xaxlab="Frequency",
  dt=1,
  xat=10^(-seq(0,4)),
  xlab=formatC(xat),
  yat=10^(-seq(0,6)),
  ygap=0,
  lty=c(1,2,3),
  dfstep=4,
  alpha=NULL,
  fcoher=NULL,
  florentz=NULL,
  alorentz=1,
  slorentz=2,
  hlor=NULL,
  fboxcar=NULL,
  portrait=TRUE,
  plotspec=TRUE,
  plotfft=TRUE,
  nwid=10,
  debug=0){
  
  if (debug > 2) browser()
  
  # consider the power spectral density of zero-mean unit variance
  
  nz <- length(tt)
  xx <- as.numeric(xx)
  reg <- !is.na(xx)
  mux <- mean(xx[reg])
  sigx <- sd(xx[reg])
  uu <- (xx-mux)/sigx
  FT <- spec.fft(uu[reg],tt[reg])
#  zz <- FT$A/nz
  zz <- FT$A
  psd <- Mod(zz)*Mod(zz) 
#  angle <- Arg(zz)
  
  if (debug > 2) browser()
  
  
  ff <- FT$fx
#  ff <- FT$fx/nz
  regpos <- FT$fx > 0
  
  if (xcal) {
    xat <- 1/c(1,12,24*c(1,7,30,183,365))
    xlab=c("1/h","2/d","1/d","1/w","1/mo","2/y","1/y")
#    xat <- 1/c(1,12,24*c(1,7,30,183,365,3650))
#    xlab=c("1/h","2/d","1/d","1/w","1/mo","2/y","1/y","1/10/y")
  } 
  
  ylab=formatC(yat)
  x <- ff[regpos]
  y <- 2*psd[regpos] # factor of 2 because... considering only positive frequencies
#  w <- angle[regpos]
  
  plot(range(xat),range(yat),log="xy",type="n",axes=FALSE,
       main=mainstr,
       xlab=xaxlab,
       ylab="Power Spectral Density")
  axis(1,at=xat,labels=xlab)
  axis(2,at=yat,labels=ylab)
  abline(h=yat,v=xat,col="gray",lty=2)
  lines(x,y)

  ### now add the annotations
  
  if (!is.null(alpha)) {
    for (i in 1:length(alpha)) {
      lines(xat,powerfun(alpha[i],xat),col="red",lty=lty[i])
    }
  } 
  
  if (!is.null(florentz)) {
    # plot the lorentzianQ
    klor <- florentz/(2*pi)
    nlor <- sum(x < klor)
    if (is.null(hlor)) {
      hlor <- pi/(florentz*nz)
    } else {
      hlor <- hlor*pi/(florentz*nz)
    }
    lines(x,lorentz(a=hlor*alorentz,k=klor,x,s=slorentz),col="red")
  }
  
  if (plotspec) {
    pspec <- pspectrum(uu[!is.na(uu)])
    lines(pspec$freq,pspec$spec/nz,col="blue",lwd=2)
  }
  
}


##################################
get_rand <- function(
  aa,
  debug=0){
  
  if (debug > 1) browser()

  bb <- aa %>%
    as_tibble() %>% # like a data.frame, but convenient for tidyverse
    mutate(zSolar   = (Solar  - mean(Solar,  na.rm=TRUE))/sd(Solar,  na.rm=TRUE), 
           zWind    = (Wind   - mean(Wind,   na.rm=TRUE))/sd(Wind,   na.rm=TRUE), 
           zDemand  = (Demand - mean(Demand, na.rm=TRUE))/sd(Demand, na.rm=TRUE)) %>% # calculate normalized variables
    group_by(HourOfWeek,MonthOfYear) %>% # group by hour of week, month of year
    mutate(pDemand = mean(zDemand, na.rm=TRUE),
           rDemand = zDemand - pDemand) %>% # add a column with the periodic demand then subtract to make random demand
    ungroup() %>%
    group_by(HourOfDay,MonthOfYear) %>% # group by how of day, month of year
    mutate(pWind = mean(zWind, na.rm=TRUE),
           rWind = zWind - pWind, 
           pSolar = mean(zSolar, na.rm=TRUE),
           rSolar = zSolar - pSolar) %>% 
    ungroup()
  
  summarize(bb)
  
  return(bb)
}

##################################
plot_rand <- function(
  aa,
  var="Solar",
  main="MISO hourly Solar",
  xat=seq(2015,2021,1),
  debug=0
  ){

  if (debug > 0) browser()
  xra <- range(xat)
  
  par(mfrow=c(3,1))

  # first, sort the values

  ss <- sort(aa$Time,index.return=TRUE)
  hh0 <- aa$Time[ss$ix] # hours
  tt0 <- hh0/366/24 # years
  xx1 <- aa$Solar[ss$ix]
  rr1 <- aa$rSolar[ss$ix]
  xx2 <- aa$Wind[ss$ix]
  rr2 <- aa$rWind[ss$ix]
  xx3 <- aa$Demand[ss$ix]
  rr3 <- aa$rDemand[ss$ix]
  
  # next, restrict analysis to the range indicated
  reg <- (tt0 <= max(xat)) & (tt0 >= min(xat))
  
  tt <- tt0[reg]
  hh <- hh0[reg]
  
  
  if (var == "Solar") {
    xx <- xx1[reg]
    rr <- rr1[reg]
  } else if (var == "Wind") {
    xx <- xx2[reg]
    rr <- rr2[reg]
  } else if (var == "Demand") {
    xx <- xx3[reg]
    rr <- rr3[reg]
  } else {
    browser()
  }

  par(mar=c(4,5,2,1))
  nf <- layout(matrix(c(1,1,2,2,3,4), nrow=3, ncol=2, byrow = TRUE), respect = FALSE); 
#  layout.show(nf)
  
  xra <- range(xat)
  yra <- range(xx[!is.na(xx)])
  plot(xra,yra,type="n",axes=FALSE,
       main=main,
       xlab="",ylab=paste(var," (MW)",sep=""))
  lines(tt,xx)
  axis(1,at=xat)
  axis(2)

  yra <- range(rr[!is.na(rr)])
  plot(xra,yra,type="n",axes=FALSE,
       main="",
       xlab="",ylab="Random (Dmnl)")
  lines(tt,rr)
  axis(1,at=xat)
  axis(2)
  
  plot_fft(xx=xx,tt=hh,xlab="",mainstr="",debug=debug,plotspec=FALSE,
           #           florentz=0.4,alorentz=2,slorentz=2.5,
           alpha=c(1,2,3))
  
  plot_fft(xx=rr,tt=hh,xlab="",mainstr="",debug=debug,plotspec=FALSE,
#           florentz=0.4,alorentz=2,slorentz=2.5,
           alpha=c(1,2,3))
  
  if (debug > 0) browser()
}


##################################
make_rand <- function(
  inputfile = "20210323_MISO_hourly.csv",
  outputfile2 = "20210323_MISO_rand.csv",
  outputfile = "MISO_periodic_random.pdf",
  timezone="Local",
  debug=0
  ){
  if (debug > 0) browser()
  
  if (debug==0) pdf(file=outputfile,height=10,width=8)
  aa <- read.csv(inputfile)
  aa <- get_rand(aa)
  write.table(aa,file=outputfile2,sep=",",row.names=FALSE)
  plot_rand(aa,var="Wind", main="MISO hourly Wind",debug=debug,xat=seq(2018.5,2021.5,0.5))
  plot_rand(aa,var="Solar", main="MISO hourly Solar",debug=debug,xat=seq(2018.5,2021.5,0.5))
  plot_rand(aa,var="Demand", main="MISO hourly Demand",debug=debug,xat=seq(2015,2022,1))
  if (debug==0) dev.off()
}